# validated: 2017-01-07 DS b9a08e826046 src/Message.cpp src/Message.h

from collections import namedtuple

from .constants import (
    kKeepAlive,
    kClientHello,
    kProtoUnsup,
    kServerHello,
    kServerHelloDone,
    kClientHelloDone,
    kEntryAssign,
    kEntryUpdate,
    kFlagsUpdate,
    kEntryDelete,
    kClearEntries,
    kExecuteRpc,
    kRpcResponse,
    
    kClearAllMagic,
    
    NT_VTYPE2RAW,
    NT_RAW2VTYPE,
)

_empty_msgtypes = [kKeepAlive, kServerHelloDone, kClientHelloDone]


MessageType = namedtuple('MessageType', ['type', 'str', 'value',
                                         'id', 'flags', 'seq_num_uid'])

class Message(object):
    
    @staticmethod
    def keepAlive():
        return MessageType(kKeepAlive, None, None, None, None, None)
    
    @staticmethod
    def clientHello(proto_rev, identity):
        return MessageType(kClientHello, identity, None, proto_rev, None, None)
    
    @staticmethod
    def protoUnsup(proto_rev):
        return MessageType(kProtoUnsup, None, None, proto_rev, None, None)
    
    @staticmethod
    def serverHelloDone():
        return MessageType(kServerHelloDone, None, None, None, None, None)
    
    @staticmethod
    def serverHello(flags, identity):
        return MessageType(kServerHello, identity, None, None, flags, None)
    
    @staticmethod
    def clientHelloDone():
        return MessageType(kClientHelloDone, None, None, None, None, None)
    
    @staticmethod
    def entryAssign(name, msg_id, seq_num_uid, value, flags):
        return MessageType(kEntryAssign, name, value, msg_id, flags, seq_num_uid)
    
    @staticmethod
    def entryUpdate(entry_id, seq_num_uid, value):
        return MessageType(kEntryUpdate, None, value, entry_id, None, seq_num_uid)
    
    @staticmethod
    def flagsUpdate(msg_id, flags):
        return MessageType(kFlagsUpdate, None, None, msg_id, flags, None)
    
    @staticmethod
    def entryDelete(entry_id):
        return MessageType(kEntryDelete, None, None, entry_id, None, None)
    
    @staticmethod
    def clearEntries():
        return MessageType(kClearEntries, None, None, kClearAllMagic, None, None)
    
    @staticmethod
    def executeRpc(rpc_id, call_uid, params):
        return MessageType(kExecuteRpc, params, None, rpc_id, None, call_uid)

    @staticmethod
    def rpcResponse(rpc_id, call_uid, result):
        return MessageType(kRpcResponse, result, None, rpc_id, None, call_uid)

    

    @staticmethod
    def read(rstream, codec, get_entry_type):
        msgtype = rstream.read(1)
        
        msg_str = None
        value = None
        msg_id = None
        flags = None
        seq_num_uid = None
        
        # switch type
        if msgtype in _empty_msgtypes:
            pass
        
        elif msgtype == kClientHello:
            msg_id, = rstream.readStruct(codec.clientHello)
            if msg_id >= 0x0300:
                msg_str = codec.read_string_v3(rstream)
        
        elif msgtype == kProtoUnsup:
            msg_id, = rstream.readStruct(codec.protoUnsup)
        
        elif msgtype == kServerHello:
            flags, = rstream.readStruct(codec.serverHello)
            msg_str = codec.read_string(rstream)
           
        elif msgtype == kEntryAssign:
            msg_str = codec.read_string(rstream)
            value_type = NT_RAW2VTYPE.get(rstream.read(1))
            if codec.proto_rev >= 0x0300:
                msg_id, seq_num_uid, flags = rstream.readStruct(codec.entryAssign)
            else:
                msg_id, seq_num_uid = rstream.readStruct(codec.entryAssign)
                flags = 0
                
            value = codec.read_value(value_type, rstream)
        
        elif msgtype == kEntryUpdate:
            if codec.proto_rev >= 0x0300:
                msg_id, seq_num_uid = rstream.readStruct(codec.entryUpdate)
                value_type = NT_RAW2VTYPE.get(rstream.read(1))
            else:
                msg_id, seq_num_uid = rstream.readStruct(codec.entryUpdate)
                value_type = get_entry_type(msg_id)
            
            value = codec.read_value(value_type, rstream)
            
        elif msgtype == kFlagsUpdate:
            msg_id, flags = rstream.readStruct(codec.flagsUpdate)
        
        elif msgtype == kEntryDelete:
            msg_id, = rstream.readStruct(codec.entryDelete)
        
        elif msgtype == kClearEntries:
            msg_id, = rstream.readStruct(codec.clearEntries)
            if msg_id != kClearAllMagic:
                raise ValueError("Bad magic")
        
        elif msgtype == kExecuteRpc:
            msg_id, seq_num_uid = rstream.readStruct(codec.executeRpc)
            msg_str = codec.read_string(rstream)
        
        elif msgtype == kRpcResponse:
            msg_id, seq_num_uid = rstream.readStruct(codec.rpcResponse)
            msg_str = codec.read_string(rstream)
            
        else:
            raise ValueError("Unrecognized message type %s" % msgtype)
        
        return MessageType(msgtype, msg_str, value, msg_id, flags, seq_num_uid)
    
    @staticmethod
    def write(msg, out, codec):
        msgtype = msg.type
        
        # switch type
        if msgtype in _empty_msgtypes:
            out.append(msgtype)
        
        elif msgtype == kClientHello:
            proto_rev = msg.id
            out += (msgtype,
                    codec.clientHello.pack(proto_rev))
            
            if proto_rev >= 0x0300:
                codec.write_string_v3(msg.str, out)
                    
        elif msgtype == kProtoUnsup:
            out += (msgtype,
                    codec.protoUnsup.pack(msg.id))
            
        elif msgtype == kServerHello:
            if codec.proto_rev >= 0x0300:
                out += (msgtype,
                        codec.serverHello.pack(msg.flags))
                codec.write_string(msg.str, out)
            
        elif msgtype == kEntryAssign:
            out.append(msgtype)
            codec.write_string(msg.str, out)
            
            value = msg.value
            if codec.proto_rev >= 0x0300:
                sb = codec.entryAssign.pack(msg.id, msg.seq_num_uid, msg.flags)
            else:
                sb = codec.entryAssign.pack(msg.id, msg.seq_num_uid)
            out += (NT_VTYPE2RAW[value.type], sb)
            
            codec.write_value(value, out)
        
        elif msgtype == kEntryUpdate:
            value = msg.value
            if codec.proto_rev >= 0x0300:
                out += (msgtype,
                        codec.entryUpdate.pack(msg.id, msg.seq_num_uid),
                        NT_VTYPE2RAW[value.type])
            else:
                out += (msgtype,
                        codec.entryUpdate.pack(msg.id, msg.seq_num_uid))
                
            codec.write_value(value, out)
        
        elif msgtype == kFlagsUpdate:
            if codec.proto_rev >= 0x0300:
                out += (msgtype,
                        codec.flagsUpdate.pack(msg.id, msg.flags))
        
        elif msgtype == kEntryDelete:
            if codec.proto_rev >= 0x0300:
                out += (msgtype,
                        codec.entryDelete.pack(msg.id))
            
        elif msgtype == kClearEntries:
            if codec.proto_rev >= 0x0300:
                out += (msgtype,
                        codec.clearEntries.pack(msg.id))
        
        elif msgtype == kExecuteRpc:
            if codec.proto_rev >= 0x0300:
                out += (msgtype,
                        codec.executeRpc.pack(msg.id, msg.seq_num_uid))
                codec.write_string(msg.str, out)
        
        elif msgtype == kRpcResponse:
            if codec.proto_rev >= 0x0300:
                out += (msgtype,
                        codec.rpcResponse.pack(msg.id, msg.seq_num_uid))
                codec.write_string(msg.str, out)
        
        else:
            raise ValueError("Internal error: bad value type %s" % msg.type)
        
        


#!/usr/bin/env python3
#
# This is a NetworkTables client (eg, the DriverStation/coprocessor side).
# You need to tell it the IP address of the NetworkTables server (the
# robot or simulator).
#
# This shows how to use a listener to listen for changes in NetworkTables
# values. This will print out any changes detected on the SmartDashboard
# table.
#

import sys
import time
from networktables import NetworkTables

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

if len(sys.argv) != 2:
    print("Error: specify an IP to connect to!")
    exit(0)

ip = sys.argv[1]

NetworkTables.initialize(server=ip)


def valueChanged(table, key, value, isNew):
    print("valueChanged: key: '%s'; value: %s; isNew: %s" % (key, value, isNew))

def connectionListener(connected, info):
    print(info, '; Connected=%s' % connected)


NetworkTables.addConnectionListener(connectionListener, immediateNotify=True)

sd = NetworkTables.getTable("SmartDashboard")
sd.addTableListener(valueChanged)

while True:
    time.sleep(1)

 



# validated: 2016-10-27 DS 86c43df src/NetworkConnection.cpp src/NetworkConnection.h
'''----------------------------------------------------------------------------'''
''' Copyright (c) FIRST 2015. All Rights Reserved.                             '''
''' Open Source Software - may be modified and shared by FRC teams. The code   '''
''' must be accompanied by the FIRST BSD license file in the root directory of '''
''' the project.                                                               '''
'''----------------------------------------------------------------------------'''

import threading

from .constants import (
    kEntryAssign,
    kEntryUpdate,
    kFlagsUpdate,
    kEntryDelete,
    kClearEntries,
)

from .message import Message
from .structs import ConnectionInfo
from .wire import WireCodec

from .support.compat import monotonic, Queue, Empty
from .support.lists import Pair
from .tcpsockets.tcp_stream import StreamEOF

import logging
logger = logging.getLogger('nt')

_empty_pair = Pair(0, 0)


class NetworkConnection(object):
    
    s_uid = 0
    s_uid_lock = threading.Lock()
    
    class State(object):
        kCreated = 0
        kInit = 1
        kHandshake = 2
        kSynchronized = 3
        kActive = 4
        kDead = 5
    
    def __init__(self, stream, notifier, handshake, get_entry_type, verbose=False):
        
        with self.s_uid_lock:
            self.m_uid = NetworkConnection.s_uid
            NetworkConnection.s_uid += 1
            
        # logging debugging
        self.m_verbose = verbose
        
        self.m_stream = stream
        self.m_notifier = notifier
        self.m_handshake = handshake
        self.m_get_entry_type = get_entry_type
        
        self.m_active = False
        self.m_proto_rev = 0x0300
        self.m_state = self.State.kCreated
        self.m_state_mutex = threading.Lock()
        self.m_last_update = 0
        
        self.m_outgoing = Queue()
        
        self.m_process_incoming = None
        self.m_read_thread = None
        self.m_write_thread = None
        
        self.m_remote_id_mutex = threading.Lock()
        self.m_remote_id = None
        self.m_last_post = 0
        
        self.m_pending_mutex = threading.Lock()
        self.m_pending_outgoing = []
        self.m_pending_update = {}
        
        # Condition variables for shutdown
        self.m_shutdown_mutex = threading.Lock()
        # Not needed in python
        #self.m_read_shutdown_cv = threading.Condition()
        #self.m_write_shutdown_cv = threading.Condition()
        self.m_read_shutdown = False
        self.m_write_shutdown = False
    
        # turn off Nagle algorithm; we bundle packets for transmission
        self.m_stream.setNoDelay()
    
    def start(self):
        if self.m_active:
            return
    
        self.m_active = True
        self.set_state(self.State.kInit)
        
        # clear queue
        try:
            while True:
                self.m_outgoing.get_nowait()
        except Empty:
            pass
    
        # reset shutdown flags
        with self.m_shutdown_mutex:
            self.m_read_shutdown = False
            self.m_write_shutdown = False
    
        # start threads
        self.m_write_thread = threading.Thread(target=self._writeThreadMain,
                                               name='nt_write_thread')
        self.m_read_thread = threading.Thread(target=self._readThreadMain,
                                               name='nt_read_thread')
        
        self.m_write_thread.daemon = True
        self.m_read_thread.daemon = True
        
        self.m_write_thread.start()
        self.m_read_thread.start()
    
    
    def stop(self):
        logger.debug("NetworkConnection stopping (%s)", self)
        
        if not self.m_active:
            return
        
        self.set_state(self.State.kDead)
        self.m_active = False
        # closing the stream so the read thread terminates
        self.m_stream.close()
    
        # send an empty outgoing message set so the write thread terminates
        self.m_outgoing.put([])
        
        # wait for threads to terminate, timeout
        self.m_write_thread.join(1)
        if self.m_write_thread.is_alive():
            logger.warn("%s did not die", self.m_write_thread.name)
            
        self.m_read_thread.join(1)
        if self.m_read_thread.is_alive():
            logger.warn("%s did not die", self.m_write_thread.name)
        
        # clear queue
        try:
            while True:
                self.m_outgoing.get_nowait()
        except Empty:
            pass
    
    def get_proto_rev(self):
        return self.m_proto_rev
    
    def get_stream(self):
        return self.m_stream
    
    def info(self):
        return ConnectionInfo(self.remote_id(), self.m_stream.getPeerIP(),
                              self.m_stream.getPeerPort(),
                              self.m_last_update, self.m_proto_rev)
        
    def is_connected(self):
        return self.m_state == self.State.kActive
    
    def last_update(self):
        return self.m_last_update
    
    def notifyIfActive(self, callback):
        with self.m_state_mutex:
            if self.m_state == self.State.kActive:
                self.m_notifier.notifyConnection(True, self.info(), callback)
    
    def set_process_incoming(self, func):
        self.m_process_incoming = func
        
    def set_proto_rev(self, proto_rev):
        self.m_proto_rev = proto_rev
    
    def set_state(self, state):
        with self.m_state_mutex:
            State = self.State
            
            # Don't update state any more once we've died
            if self.m_state == State.kDead:
                return
            
            # One-shot notify state changes
            if self.m_state != State.kActive and state == State.kActive:
                info = self.info()
                self.m_notifier.notifyConnection(True, info)
                logger.info("CONNECTED %s port %s (%s)",
                            info.remote_ip, info.remote_port, info.remote_id)
            elif self.m_state != State.kDead and state == State.kDead:
                info = self.info()
                self.m_notifier.notifyConnection(False, info)
                logger.info("DISCONNECTED %s port %s (%s)",
                            info.remote_ip, info.remote_port, info.remote_id)
                
        self.m_state = state
    
    def state(self):
        return self.m_state
    
    def remote_id(self):
        with self.m_remote_id_mutex:
            return self.m_remote_id
    
    def set_remote_id(self, remote_id):
        with self.m_remote_id_mutex:
            self.m_remote_id = remote_id
            
    def uid(self):
        return self.m_uid
     
    def _sendMessages(self, msgs):
        self.m_outgoing.put(msgs)
    
    def _readThreadMain(self):
        decoder = WireCodec(self.m_proto_rev)
        
        verbose = self.m_verbose
        
        def _getMessage():
            decoder.set_proto_rev(self.m_proto_rev)
            try:
                return Message.read(self.m_stream, decoder, self.m_get_entry_type)
            except IOError as e:
                logger.warn("read error in handshake: %s", e)
                
                # terminate connection on bad message
                self.m_stream.close()
                
                return None
            
        self.set_state(self.State.kHandshake)
        
        try:
            handshake_success = self.m_handshake(self, _getMessage, self._sendMessages)
        except Exception:
            logger.exception("Unhandled exception during handshake")
            handshake_success = False
        
        if not handshake_success:
            self.set_state(self.State.kDead)
            self.m_active = False
        else:
            self.set_state(self.State.kActive)
            
            try:
                while self.m_active:
                    if not self.m_stream:
                        break
            
                    decoder.set_proto_rev(self.m_proto_rev)
                    
                    try:
                        msg = Message.read(self.m_stream, decoder, self.m_get_entry_type)
                    except Exception as e:
                        if not isinstance(e, StreamEOF):
                            if verbose:
                                logger.exception("read error")
                            else:
                                logger.warn("read error: %s", e)
                        
                        # terminate connection on bad message
                        self.m_stream.close()
            
                        break
            
                    if verbose:
                        logger.debug('received type=%s with str=%s id=%s seq_num=%s value=%s',
                                     msg.type, msg.str, msg.id, msg.seq_num_uid, msg.value)
                    
                    self.m_last_update = monotonic()
                    self.m_process_incoming(msg, self)
            except IOError as e:
                # connection died probably
                logger.debug("IOError in read thread: %s", e)
            except Exception:
                logger.warn("Unhandled exception in read thread", exc_info=True)
                
            logger.debug("read thread died (%s)", self)
            self.set_state(self.State.kDead)
            self.m_active = False
        
        # also kill write thread
        self.m_outgoing.put([])  
        
        with self.m_shutdown_mutex:
            self.m_read_shutdown = True
    
    def _writeThreadMain(self):
        encoder = WireCodec(self.m_proto_rev)
    
        verbose = self.m_verbose
        out = []
    
        try:
            while self.m_active:
                msgs = self.m_outgoing.get()
                
                if verbose:
                    logger.debug("write thread woke up")
                
                if not msgs:
                    continue
        
                encoder.set_proto_rev(self.m_proto_rev)
                
                if verbose:
                    logger.debug('sending %s messages', len(msgs))
                
                for msg in msgs:
                    if msg:
                        if verbose:
                            logger.debug('sending type=%s with str=%s id=%s seq_num=%s value=%s',
                                         msg.type, msg.str, msg.id, msg.seq_num_uid, msg.value)
                        
                        Message.write(msg, out, encoder)
                
                if not self.m_stream:
                    break
        
                if not out:
                    continue
                
                self.m_stream.send(b''.join(out))
                
                del out[:]
        
                #if verbose:
                #    logger.debug('send %s bytes', encoder.size())
        except IOError as e:
            # connection died probably
            if not isinstance(e, StreamEOF):
                logger.debug("IOError in write thread: %s", e)
        except Exception:
            logger.warn("Unhandled exception in write thread", exc_info=True)
            
    
        logger.debug('write thread died (%s)', self)
        self.set_state(self.State.kDead)
        self.m_active = False
        self.m_stream.close() # also kill read thread
        
        with self.m_shutdown_mutex:
            self.m_write_shutdown = True
    
    def queueOutgoing(self, msg):
        with self.m_pending_mutex:
    
            # Merge with previous.  One case we don't combine: delete/assign loop.
            msgtype = msg.type
            if msgtype in [kEntryAssign, kEntryUpdate]:
            
                # don't do this for unassigned id's
                msg_id = msg.id
                if msg_id == 0xffff:
                    self.m_pending_outgoing.append(msg)
                    return
        
                mpend = self.m_pending_update.get(msg_id)
                if mpend is not None and mpend.first != 0:
                    # overwrite the previous one for this id
                    oldidx = mpend.first - 1
                    oldmsg = self.m_pending_outgoing[oldidx]
                    if (oldmsg and oldmsg.type == kEntryAssign and msgtype == kEntryUpdate):
                        # need to update assignment with seq_num and value
                        oldmsg = Message.entryAssign(oldmsg.str, msg_id, msg.seq_num_uid,
                                                      msg.value, oldmsg.flags)
        
                    else:
                        oldmsg = msg # easy update
                        
                    self.m_pending_outgoing[oldidx] = oldmsg
                        
                else:
                    # new, remember it
                    pos = len(self.m_pending_outgoing)
                    self.m_pending_outgoing.append(msg)
                    self.m_pending_update[msg_id] = Pair(pos + 1, 0)

            elif msgtype == kEntryDelete:
                # don't do this for unassigned id's
                msg_id = msg.id
                if msg_id == 0xffff:
                    self.m_pending_outgoing.append(msg)
                    return
                
                # clear previous updates
                mpend = self.m_pending_update.get(msg_id)
                if mpend is not None:
                    if mpend.first != 0:
                        self.m_pending_outgoing[mpend.first - 1] = None
        
                    if mpend.second != 0:
                        self.m_pending_outgoing[mpend.second - 1] = None
                        
                    self.m_pending_update[msg_id] = _empty_pair
                
                # add deletion
                self.m_pending_outgoing.append(msg)
                
            elif msgtype == kFlagsUpdate:
                # don't do this for unassigned id's
                msg_id = msg.id
                if id == 0xffff:
                    self.m_pending_outgoing.append(msg)
                    return
        
                mpend = self.m_pending_update.get(msg_id)
                if mpend is not None and mpend.second != 0:
                    # overwrite the previous one for this id
                    self.m_pending_outgoing[mpend.second - 1] = msg
        
                else:
                    # new, remember it
                    pos = len(self.m_pending_outgoing)
                    self.m_pending_outgoing.append(msg)
                    self.m_pending_update[msg_id] = Pair(0, pos + 1)
            
            elif msgtype == kClearEntries:
                # knock out all previous assigns/updates!
                for i, m in enumerate(self.m_pending_outgoing):
                    if not m:
                        continue
        
                    t = m.type
                    if t in [kEntryAssign, kEntryUpdate, kFlagsUpdate,
                             kEntryDelete, kClearEntries]:
                        self.m_pending_outgoing[i] = None
        
                self.m_pending_update.clear()
                self.m_pending_outgoing.append(msg)
        
            else:
                self.m_pending_outgoing.append(msg)
    
    def postOutgoing(self, keep_alive):
        with self.m_pending_mutex:
            now = monotonic()
            if not self.m_pending_outgoing:
                if not keep_alive:
                    return
        
                # send keep-alives once a second (if no other messages have been sent)
                if (now - self.m_last_post) < 1.0:
                    return
        
                self.m_outgoing.put((Message.keepAlive(),))
        
            else:
                self.m_outgoing.put(self.m_pending_outgoing)
                
                self.m_pending_outgoing = []
                self.m_pending_update.clear()
        
            self.m_last_post = now
    



# notrack

from collections import namedtuple

Pair = namedtuple('Pair', ['first', 'second'])

def ensure_id_exists(lst, msg_id, default=None):
    if msg_id >= len(lst):
        lst += [default]*(msg_id - len(lst)+1)





# This is imported first to avoid circular dependency problems
try:
    from .version import __version__
except ImportError:
    __version__ = 'master'

from .networktables import NetworkTables

# Deprecated, will be removed in 2018
from .networktable import NetworkTable





# validated: 2016-10-23 DS c2ae897 wpiutil/src/support/leb128.cpp
'''
    Operations related to LEB128 encoding/decoding

    The algorithm is taken from Appendix C of the DWARF 3 spec. For information
    on the encodings refer to section "7.6 - Variable Length Data"

'''

import sys

def size_uleb128(value):
    count = 0
    while True:
        value >>= 7
        count += 1
        if value == 0:
            break
    return count



# You can do bitwise operations on bytes, not so with strings (py2)
if sys.version_info[0] == 2:
    
    def encode_uleb128(value):
        out = bytearray()
        while True:
            byte = value & 0x7F
            value >>= 7
            if value != 0:
                byte = byte | 0x80
            out.append(byte)
            if value == 0:
                break
        return bytes(out)
    
    def read_uleb128(rstream):
        result = 0
        shift = 0
        while True:
            x = rstream.read(1)
            b = ord(x)
            result |= (b & 0x7f) << shift
            shift += 7
            if (b & 0x80) == 0:
                break
        return result
    
else:
    
    def encode_uleb128(value):
        out = bytearray()
        while True:
            byte = value & 0x7F
            value >>= 7
            if value != 0:
                byte = byte | 0x80
            out.append(byte)
            if value == 0:
                break
        return out
    
    def read_uleb128(rstream):
        result = 0
        shift = 0
        while True:
            b = rstream.read(1)[0]
            result |= (b & 0x7f) << shift
            shift += 7
            if (b & 0x80) == 0:
                break
        return result




#!/usr/bin/env python3
#
# Simple benchmark script that tests NetworkTables throughput.
#
# To run the server
#
#     python3 benchmark.py
# 
# To run the client:
#
#    python3 benchmark.py 127.0.0.1
#
# In theory, the limiting factor is internal buffering and not network
# bandwidth, so running this test on a local machine should be mostly
# valid.
#
# On OSX with default write flush period, I average ~18hz
#

from __future__ import print_function

import sys
import time

import logging
logging.basicConfig(level=logging.DEBUG)

from networktables import NetworkTables

class Benchmark(object):

    def __init__(self):
        
        client = False
        
        if len(sys.argv) > 1:
            client = True
            NetworkTables.initialize(server=sys.argv[1])
         
        # Default write flush is 0.05, could adjust for less latency   
        #NetworkTable.setWriteFlushPeriod(0.01)
        
        
        self.nt = NetworkTables.getTable('/benchmark')
        self.updates = 0
        
        if client:
            self.nt.addTableListener(self.on_update)
            self.recv_benchmark()
        else:
            self.send_benchmark()

    def on_update(self, *args):
        self.updates += 1

    def recv_benchmark(self):
        
        print("Starting to receive")
        
        last = None
        last_updates = 0
        
        while True:
            
            now = time.time()
            updates = self.updates
            
            if last is not None:
                rate = (updates - last_updates) / (now - last)
                print('Update rate:', rate)
            
            last = now
            last_updates = updates
            time.sleep(1)
            
    def send_benchmark(self):
        
        print("Sending")
        
        i = 0
        
        while True:
            i += 1
            self.nt.putNumber('key', i)
            time.sleep(0.0001)
    
if __name__ == '__main__':
    Benchmark()